/**
 * @brief 一个简单通过协议进行双边报价报撤单的功能演示
 *
 * 演示功能：
 * 1. 通过配置文件创建API实例；
 * 2. 启动API，启动成功后，调用login接口登录柜台；
 * 3. 等数据加载完毕后，使用协议方式发送双边报价报单请求；
 * 4. 等待3秒后，针对所有的报价单，使用协议方式尝试撤单；
 * 5. 等待3秒后，调用logout接口登出柜台；
 */

#include <fcntl.h>
#include "unistd.h"
#if defined(_WIN32) || defined(WIN32)
#include <winsock2.h>
#include <ws2tcpip.h>
#else
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#endif

#include <map>
#include "ExampleTrader.h"

#pragma pack(push, 1)

struct CXeleFairInputQuoteMsg {
    uint8_t                 messageId;                      ///< 消息id, 报单操作为0x6F
    uint8_t                 clientIndex;                    ///< 客户端index
    uint32_t                token;                          ///< 客户端token
    int                     seqNo;                          ///< 消息序列号
    uint32_t                requestID;                      ///< 请求id，建议单调递增，有效范围为(0~0xfffffffe)
    int                     localOrderNo;                   ///< 报单本地id
    uint32_t                instrumentIndex;                ///< 合约序号
    uint8_t                 bidOffsetFlag;                  ///< 买方开平标记
    uint8_t                 bidHedgeFlag;                   ///< 买方投保标记
    uint16_t                bidVolume;                      ///< 买入数量
    double                  bidPrice;                       ///< 买方价格
    uint8_t                 askOffsetFlag;                  ///< 卖方开平标记
    uint8_t                 askHedgeFlag;                   ///< 卖方投保标记
    uint16_t                askVolume;                      ///< 卖入数量
    double                  askPrice;                       ///< 卖方价格
    uint64_t                quoteDemandNo;                  ///< 询价编号
    uint8_t                 exchangeFront;                  ///< 前置信息 不指定前置填写为0，指定前置需要加上偏移量10，如指定前置3，则需要填写13
    uint32_t                userRef;                        ///< 用户自定义数据
    char                    reserve[5];                     ///< 预留字段
};
static_assert(sizeof(CXeleFairInputQuoteMsg) == 64, "CXeleFairInputQuoteMsg size error");

struct CXeleFairQuoteActionMsg{
    uint8_t                 messageId;                      ///< 消息id, 撤单操作为0x71
    uint8_t                 clientIndex;                    ///< 客户端index
    uint32_t                token;                          ///< 客户端token
    int                     seqNo;                          ///< 消息序列号
    uint32_t                requestID;                      ///< 请求id，建议单调递增，有效范围为(0~0xfffffffe)
    int                     localActionNo;                  ///< 本地操作编号
    int                     systemOrderNo;                  ///< 被撤单柜台编码
    char                    actionFlag;                     ///< 报单操作标志
    int                     localOrderNo;                   ///< 本地报单编号
    uint8_t                 exchangeFront;                  ///< 前置信息 不指定前置填写为0，指定前置需要加上偏移量10，如指定前置3，则需要填写13
    uint32_t                userRef;                        ///< 用户自定义数据
    char                    reserve[32];                    ///< 预留字段
};
static_assert(sizeof(CXeleFairQuoteActionMsg) == 64, "CXeleFairQuoteActionMsg size error");

struct CXeleFairQuoteDemandMsg {
    uint8_t                 messageId;                      ///< 消息id，询价操作为0x73
    uint8_t                 clientIndex;                    ///< 客户端index
    uint32_t                token;                          ///< 客户端token
    int                     seqNo;                          ///< 消息序列号
    int                     requestID;                      ///< 请求id，建议单调递增，有效范围为(0~0xfffffffe)
    int                     localOrderNo;                   ///< 询价本地id
    char                    instrumentID[16];               ///< 合约代码
    uint32_t                instrumentIndex;                ///< 合约序号
    uint8_t                 exchangeFront;                  ///< 前置信息 不指定前置填写为0，指定前置需要加上偏移量10，如指定前置3，则需要填写13
    uint32_t                userRef;                        ///< 用户自定义数据
    char                    reserve[21];                    ///< 预留字段
};
static_assert(sizeof(CXeleFairQuoteDemandMsg) == 64, "CXeleFairQuoteDemandMsg size error");

#pragma pack(pop)


class Example_02_Trader : public ExampleTrader {
public:
    Example_02_Trader() = default;
    ~Example_02_Trader() = default;

    void onStart(int errorCode, bool isFirstTime) override {
        ExampleTrader::onStart(errorCode, isFirstTime);

        printf("api start ok.\n");
    }

    void onStop(int reason) override {
        ExampleTrader::onStop(reason);

        printf("api stopped, reason: %d.\n", reason);
    }

    void onServerReboot() override {
        printf("trade system reboot, clear local data.\n");
        mInstrument = nullptr;
        mOrders.clear();
    }

    void onLogin(int errorCode, int exchangeCount) override {
        ExampleTrader::onLogin(errorCode, exchangeCount);

        if (errorCode != 0) {
            printf("login failed, error code: %d.\n", errorCode);
        } else {
            printf("login success, exchange count=%d.\n", exchangeCount);
        }
    }

    void onLogout(int errorCode) override {
        ExampleTrader::onLogout(errorCode);

        printf("logout success.\n");
    }

    void onReadyForTrading(const XTFAccount *account) override {
        ExampleTrader::onReadyForTrading(account);

        printf("ready for trading.\n");
        mInstrument = mApi->getInstrumentByID(mInstrumentID.c_str()); // 查询可用的合约对象
        if (!mInstrument) {
            printf("instrument not found: %s\n", mInstrumentID.c_str());
            exit(0);
        }

        mExchange = mInstrument->getProduct()->getProductGroup()->getExchange();
        if (!mExchange) {
            printf("instrument error, exchange invalid: %s\n", mInstrumentID.c_str());
            exit(0);
        }

        mLocalOrderID = account->lastLocalOrderID;   // API&协议报单使用
        mLocalActionID = account->lastLocalActionID; // 协议报单使用
    }

    void onLoadFinished(const XTFAccount *account) override {
        ExampleTrader::onLoadFinished(account);

        printf("load data finished.\n");
    }

    void onAccount(int event, int action, const XTFAccount *account) override {
        // 账户信息发生变化时回调该接口，如：出入金变化
        if (event == XTF_EVT_AccountCashInOut) {
            if (action == XTF_CASH_In) printf("cash in.\n");
            if (action == XTF_CASH_Out) printf("cash out.\n");
        } else {
            printf("account is changed: event=%d, action=%d, accountID=%s\n",
                   event, action, account->accountID);
        }
    }

    void onExchange(int event, int channelID, const XTFExchange *exchange) override {
        // 交易所信息发生变化时回调该接口，如：交易所前置变化
        printf("exchange is changed: event=%d, channelID=%d, exchange=%s\n",
               event, channelID, exchange->exchangeID);
    }

    void onInstrument(int event, const XTFInstrument *instrument) override {
        // 合约属性发生变化时回调该接口，如：状态变化
        if (event == XTF_EVT_InstrumentStatusChanged) {
            printf("instrument status changed: %s %d.\n",
                   instrument->instrumentID, instrument->status);
        }
    }

    void onChangePassword(int errorCode) override {
        printf("password is changed: errorCode=%d\n", errorCode);
    }

    void onOrder(int errorCode, const XTFOrder *order) override {
        printf("recv order report: action=%d, sys-id=%d, "
               "status=%s, error-code=%d.\n",
               order->actionType, order->sysOrderID,
               getOrderStatus(order->orderStatus), errorCode);
        if (errorCode == 0) {
            if (order->orderStatus == XTF_OS_Queuing) {
                mOrders[order->sysOrderID] = order;
            }
        }
    }

    void onCancelOrder(int errorCode, const XTFOrder *cancelOrder) override {
        printf("recv cancel order report: sys-id=%d, status=%s, error-code=%d.\n",
               cancelOrder->sysOrderID, getOrderStatus(cancelOrder->orderStatus), errorCode);
        if (errorCode == 0 || errorCode == 1198) { // 1198: 报单已经撤销或成交或为错单
            auto iter = mOrders.find(cancelOrder->sysOrderID);
            if (iter != mOrders.end())
                mOrders.erase(iter);
        }
    }

    void onTrade(const XTFTrade *trade) override {
        printf("recv trade report: trade-id=%ld, price=%.4f, volume=%d/%d, sys-order-id=%d\n",
                trade->tradeID, trade->tradePrice, trade->order->totalTradedVolume,
                trade->order->orderVolume, trade->order->sysOrderID);
    }

    void onEvent(const XTFEvent &event) override {
        printf("recv event: %d.\n", event.eventID);
    }

    void onQuote(int errorCode, const XTFQuote *quote) override {
        printf("recv quote report: action=%d, sys-id=%d, "
               "status=%s, error-code=%d.\n",
               quote->actionType, quote->sysOrderID,
               getOrderStatus(quote->orderStatus), errorCode);
        if (errorCode == 0) {
            if (quote->orderStatus == XTF_OS_Queuing) {
                mQuotes[quote->sysOrderID] = quote;
            }
        }
    }

    void onCancelQuote(int errorCode, const XTFQuote *quote) override {
        printf("recv cancel quote report: sys-id=%d, status=%s, error-code=%d.\n",
               quote->sysOrderID, getOrderStatus(quote->orderStatus), errorCode);
        if (errorCode == 0) {
            auto iter = mQuotes.find(quote->sysOrderID);
            if (iter != mQuotes.end())
                mQuotes.erase(iter);
        }
    }

    void onDemandQuote(int errorCode, const XTFInputQuoteDemand &inputQuoteDemand) override {
        // TODO: 询价请求的处理结果，如果 ErrorCode 非0，则表示询价失败；否则，询价成功
        printf("demand quote result: %d, localOrderID=%u, instrumentID=%s, userRef=%u\n",
               errorCode, inputQuoteDemand.localOrderID,
               inputQuoteDemand.instrument->instrumentID,
               inputQuoteDemand.userRef);
    }

    void onQuoteDemandNotify(const XTFQuoteDemandNotify &notify) override {
        // TODO: 做市商收到询价通知后，可以在此接口发送双边报价
        printf("quote demand notify: %s, quoteDemandID=%lu, demandDay=%s, demandTime=%s\n",
               notify.instrument->instrumentID, notify.quoteDemandID,
               notify.demandDay, notify.demandTime);
    }

    void onError(int errorCode, void *data, size_t size) override {
        printf("something is wrong, error code: %d.\n", errorCode);
    }

    void start() {
        if (mApi) {
            printf("error: trader has been started.\n");
            return;
        }

        mLocalOrderID = 0;
        mApi = makeXTFApi(mConfigPath.c_str());
        if (mApi == nullptr) {
            printf("error: create xtf api failed, please check config: %s.\n", mConfigPath.c_str());
            exit(0);
        }

        // 保存报单的地址和端口，用于协议方式发送报撤单。
        mTradeServerIp = mApi->getConfig("TRADE_SERVER_IP");
        mTradeServerPort = mApi->getConfig("TRADE_SERVER_PORT");
        openUdpSocket();

        printf("api starting..., config: %s.\n", mConfigPath.c_str());
        int ret = mApi->start(this);
        if (ret != 0) {
            printf("start failed, error code: %d\n", ret);
            exit(0);
        }
    }

    void stop() {
        if (!mApi) {
            printf("error: trader is not started.\n");
            return;
        }

        closeUdpSocket();

        printf("api stopping...\n");
        int ret = mApi->stop();
        if (ret == 0) {
            // API停止操作是异步操作，需要等待一定时间，以防API对象回调时失效。
            // 4.1.664及更高版本不存在此问题，不需要增加延时。
            Sleep(1000);

            delete mApi;
            mApi = nullptr;
        } else {
            printf("api stop failed, error code: %d\n", ret);
        }
    }

    void login() {
        if (!mApi) return;
        printf("api logging in...\n");
        int ret = mApi->login();
        if (ret != 0) {
            printf("api logging in failed, error code: %d\n", ret);
        }
    }

    void logout() {
        if (!mApi) return;
        printf("api logging out...\n");
        int ret = mApi->logout();
        if (ret != 0) {
            printf("api logging out failed, error code: %d\n", ret);
        }
    }

    // 协议方式报单
    void insertQuoteByUdp() {
        if (!mApi) {
            printf("api is not started.\n");
            return;
        }

        if (!mInstrument) {
            printf("instrument is not found: %s\n", mInstrumentID.c_str());
            return;
        }

        if (!mExchange) {
            printf("exchange invalid\n");
            return;
        }

        printf("udp prepare quote...\n");
        CXeleFairInputQuoteMsg quote{};
        quote.messageId = 0x6F;
        quote.clientIndex = mExchange->clientIndex;
        quote.token = mExchange->clientToken;
        quote.localOrderNo = ++mLocalOrderID;
        quote.requestID = ++mRequestID;
        quote.seqNo = ++mSeqNo;
        quote.instrumentIndex = mInstrument->instrumentIndex;
        quote.bidOffsetFlag = XTF_OF_Open;
        quote.bidHedgeFlag = XTF_HF_Speculation;
        quote.bidVolume = (uint16_t) mVolume;
        quote.bidPrice = mPrice;
        quote.askOffsetFlag = XTF_OF_Open;
        quote.askHedgeFlag = XTF_HF_Speculation;
        quote.askVolume = (uint16_t) mVolume;
        quote.askPrice = mPrice + 1.0; // 卖价不能小于等于买价
        quote.quoteDemandNo = 0; // 根据需要填写
        quote.exchangeFront = 0;
        quote.userRef = 0; // 根据需要填写

        printf("udp insert quote...\n");
        sendUdpData(&quote, sizeof(quote));
    }

    // 协议方式撤单
    void cancelQuoteByUdp() {
        if (!mApi) {
            printf("api is not started.\n");
            return;
        }

        if (mQuotes.empty()) {
            printf("no orders need cancel.\n");
            return;
        }

        if (!mExchange) {
            printf("exchange invalid\n");
            return;
        }

        printf("udp cancel quote...\n");
        auto quotes = mQuotes;
        printf("quote count: %lu\n", quotes.size());
        for (auto &iter: quotes) {
            printf("cancel quote: sys-id=%d.\n", iter.first);
            CXeleFairQuoteActionMsg quoteAction{};
            quoteAction.messageId = 0x71;
            quoteAction.clientIndex = mExchange->clientIndex;
            quoteAction.token = mExchange->clientToken;
            quoteAction.seqNo = ++mSeqNo;
            quoteAction.requestID = ++mRequestID;
            quoteAction.localActionNo = ++mLocalActionID;
            quoteAction.systemOrderNo = iter.first;
            quoteAction.actionFlag = '0';
            quoteAction.localOrderNo = iter.second->localOrderID;
            quoteAction.userRef = 0; // 根据需要填写
            sendUdpData(&quoteAction, sizeof(quoteAction));
            Sleep(500); // sleep 500ms;
        }
    }

    // 协议方式撤单
    void demandQuoteByUdp() {
        if (!mApi) {
            printf("api is not started.\n");
            return;
        }

        if (!mExchange) {
            printf("exchange invalid\n");
            return;
        }

        printf("udp demand quote...\n");
        CXeleFairQuoteDemandMsg quoteDemand{};
        quoteDemand.messageId = 0x73;
        quoteDemand.clientIndex = mExchange->clientIndex;
        quoteDemand.token = mExchange->clientToken;
        quoteDemand.seqNo = ++mSeqNo;
        quoteDemand.requestID = ++mRequestID;
        quoteDemand.localOrderNo = ++mLocalOrderID;
        strncpy(quoteDemand.instrumentID, mInstrument->instrumentID, sizeof(quoteDemand.instrumentID));
        quoteDemand.instrumentIndex = mInstrument->instrumentIndex;
        quoteDemand.exchangeFront = 0;
        quoteDemand.userRef = 0; // 根据需要填写
        sendUdpData(&quoteDemand, sizeof(quoteDemand));
    }

    void openUdpSocket() {
        if (mUdpSocket != -1) return;
        mUdpSocket = ::socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
        if (mUdpSocket == -1) {
            printf("udp create socket failed: %d\n", errno);
            exit(1);
        }
        printf("udp channel create socket success: %d\n", mUdpSocket);

        std::string remoteIp = mTradeServerIp;
        uint16_t remotePort = (uint16_t) std::stoi(mTradeServerPort);
        struct sockaddr_in sa{};
        memset(&sa, 0, sizeof(sa));
        sa.sin_family = AF_INET;
        sa.sin_addr.s_addr = inet_addr(remoteIp.c_str());
        sa.sin_port = htons(remotePort);

        int result = connect(mUdpSocket, (struct sockaddr *) &sa, sizeof(sockaddr_in));
        if (result == -1) {
            printf("udp channel connect[%s:%u] failed, errno: %d.", remoteIp.c_str(), remotePort, errno);
            ::_close(mUdpSocket); // 释放创建的套接字
            mUdpSocket = -1;
            exit(2);
        }
    }

    void closeUdpSocket() {
        if (mUdpSocket == -1) return;
        ::_close(mUdpSocket);
        mUdpSocket = -1;
    }

    void sendUdpData(const void *data, size_t len) {
#if defined(_WIN32) || defined(WIN32)
        int result = ::send(mUdpSocket, (const char *)data, len, 0);
#else
        int result = ::send(mUdpSocket, data, len, 0);
#endif
        if (result == (int) len) {
            printf("udp send ok, data size: %d\n", result);
        } else {
            printf("udp send error: %d\n", result);
        }
    }

    int sendWarmUp() {
        if (!mApi) return -1; // api not created.
        char warmOrderBuf[64];
        int ret = mApi->buildWarmOrder(warmOrderBuf, 64);
        if (ret != 0) {
            printf("build warm order failed, error code: %d\n", ret);
            return ret; // build warm order failed.
        }
        sendUdpData(warmOrderBuf, 64);
        return 0;
    }

private:
    const XTFExchange *mExchange;
    const XTFInstrument *mInstrument;
    std::map<int, const XTFOrder *> mOrders;
    std::map<int, const XTFQuote *> mQuotes;
    int mLocalOrderID = 0;
    int mLocalActionID = 0;
    int mUdpSocket = -1;
    std::string mTradeServerIp;
    std::string mTradeServerPort;
    int mSeqNo = 0;
    uint32_t mRequestID = 0;
};


/**
 * @brief 一个简单的协议报撤单功能演示
 *
 * API登录柜台后，使用协议方式报一手多头开仓单，等待3秒后，使用协议方式尝试撤单。
 *
 * @param configPath
 * @param instrumentId
 * @param price
 * @param volume
 */
void runExample(const std::string &configPath, const std::string &instrumentId, double price, int volume) {
    printf("start example 03.\n");

    Example_02_Trader trader;
    trader.setConfigPath(configPath);
    trader.setInstrumentID(instrumentId);
    trader.setPrice(price);
    trader.setVolume(volume);

    trader.start();
    while (!trader.isStarted())
        trader.wait(1, "wait for trader started");

    trader.login();
    while (!trader.isLoadFinished())
        trader.wait(1, "wait for data load finished");

    trader.insertQuoteByUdp();
    trader.wait(3, "wait for quote inserted");

    trader.cancelQuoteByUdp();
    trader.wait(3, "wait for quote canceled");

    trader.logout();
    while (!trader.isLoggedOut())
        trader.wait(1, "wait for trader logout");

    trader.stop();
    while (!trader.isStopped())
        trader.wait(1, "wait for trader stopped");
}

int main(int argc, const char *argv[]) {
    printf("api version: %s.\n", getXTFVersion());

    // TODO: 解析传入参数，提取相关的配置
    std::string configPath = "xtf_trader_api.config";
    std::string instrumentId = "au2312";
    double price = 420.0;
    int volume = 1;
    runExample(configPath, instrumentId, price, volume);
    return 0;
}
